from pathlib import Path
import polars as pl
import re
from shutil import rmtree
from sklearn.feature_extraction.text import TfidfVectorizer

# Pre-compile the regular expression
regex_pattern = re.compile("[\(\[\<\"].*?[\)\]\>\"]")


class PipelineText:
    @staticmethod
    def clean_text(data: pl.DataFrame, col: str = 'item_name') -> pl.DataFrame:
        # Use the pre-compiled regex pattern
        def clean_func(x):
            return regex_pattern.sub("", x).lower().rstrip('.').strip()

        return data.with_columns(
            pl.col(col).map_elements(clean_func, return_dtype=pl.String)
            .alias(f'{col.lower()}_clean')
        )


def ngrams_func(string: str):
    ngrams = zip(*[string[i:] for i in range(3)])
    return [''.join(ngram) for ngram in ngrams]


def tfidf(lst_item: list, dim: int = 512):
    # Ensure the vectorizer is as efficient as possible
    vectorizer = TfidfVectorizer(analyzer=ngrams_func, max_features=dim)
    vectorizer.fit(lst_item)
    return vectorizer


def rm_all_folder(path: Path) -> None:
    # Use shutil.rmtree for more efficient directory removal
    rmtree(path)


def make_dir(folder_name: Path) -> None:
    # Remove redundant existence check
    folder_name.mkdir(parents=True, exist_ok=True)



from pathlib import Path
import pandas as pd
import polars as pl
from core_pro.ultilities import make_dir
import sys
import subprocess
from loguru import logger
from tqdm import tqdm
import orjson
from .func import PipelineText
import duckdb

logger.remove()
logger.add(sys.stdout, colorize=True, format='<green>{time:HH:mm:ss}</green> | <level>{level}</level> | <cyan>{function}</cyan> | <level>{message}</level>')


class PipelineImage:
    def __init__(self, path: Path, col_image: str = 'image_url'):
        self.path_image = path / 'download_img'
        self.col_image = col_image

        # Initialize path for images
        make_dir(self.path_image)

    def download_images(self, file_name: str):
        command = [
            "img2dataset",
            f"--url_list={self.path_image / (file_name + '_0.parquet')}",
            f"--output_folder={self.path_image / ('img_' + file_name)}",
            "--processes_count=16",
            "--thread_count=32",
            "--image_size=224",
            "--output_format=files",
            "--input_format=parquet",
            f"--url_col={self.col_image}",
            "--number_sample_per_shard=50000",
        ]
        subprocess.run(command, cwd=self.path_image)

    def load_images(self, mode: str = '') -> pl.DataFrame:
        # Listing
        path = self.path_image / f'img_{mode}'
        lst_json = sorted(path.glob('*/*.json'))
        lst_file = [orjson.loads(open(str(i), "r").read())['url'] for i in tqdm(lst_json, desc='Loading json in folder')]
        lst_img = [str(i) for i in tqdm(sorted(path.glob('*/*.jpg')), desc='Loading jpg in folder')]

        df = pl.DataFrame({
            f'{mode}_{self.col_image}': lst_file,
            f'{mode}_file_path': lst_img,
            f'{mode}_exists': [True] * len(lst_file),
        })

        logger.info(f'[Data] Load Images: {df.shape}')
        return df

    def run(self, data: pl.DataFrame | pd.DataFrame, mode: str = '', download: bool = False):
        # Load data
        query = f"""
        SELECT *
        , CONCAT('http://f.shopee.vn/file/', UNNEST(array_slice(string_split(images, ','), 1, 1))) AS {self.col_image}
        FROM data
        """
        df = duckdb.query(query).pl()

        logger.info(f'[Data] Base Data: {df.shape}')

        # Download
        if download:
            df.write_parquet(self.path_image / f'{mode}_0.parquet')
            self.download_images(mode)

        # Load image data
        data_img = self.load_images(mode)

        # Join
        data = (
            df.drop(['images'])
            .pipe(PipelineText.clean_text)
            .select(pl.all().name.prefix(f'{mode}_'))
            .join(data_img, on=f'{mode}_{self.col_image}', how='left')
            .filter(pl.col(f'{mode}_exists'))
        )

        logger.info(f'[Data] Join Images: {data.shape}')
        return data, data_img



from pathlib import Path
import duckdb
import polars as pl
from time import perf_counter
from .build_index.func import rm_all_folder, PipelineText, make_dir
from .build_index.matching import BELargeScale


class Matching:
    def __init__(
            self,
            col_category: str,
            path: Path,
            path_database: Path = None,
            path_query: Path = None,
            query_batch_size: int = 500_000,
            match_mode: str = 'text_dense',
    ):
        self.path = path
        self.path_database = path_database
        self.path_query = path_query
        self.col_category = col_category
        self.query_batch_size = query_batch_size
        self.match_mode = match_mode

    def check_file_type(
            self,
            file_path: Path,
            mode: str = '',
    ) -> dict:
        # check path
        file_type = file_path.suffix[1:]

        # read duckdb
        query = f"""select * from read_{file_type}('{file_path}')"""
        df = duckdb.sql(query).pl()

        # clean data
        if self.match_mode != 'image':
            df = (
                df
                .pipe(PipelineText.clean_text)
                .select(pl.all().name.prefix(f'{mode}_'))
                .drop_nulls()
            )

        # export
        file_path = self.path / f'{mode}_{self.match_mode}_clean.parquet'
        df.write_parquet(file_path)

        # status
        return {
            f'{mode}_file_clean_path': file_path,
            f'{mode}_data_shape': df.shape[0],
            f'{mode}_col_category': sorted(df[f'{mode}_{self.col_category}'].unique())
        }

    def run(
            self,
            export_type: str = 'parquet',
            top_k: int = 10,
    ):
        """
        Run matching processes
        :param export_type: parquet | csv
        :param top_k: top k matches
        :return: json
        """
        # init
        json_stats = {}

        # read file
        status = self.check_file_type(file_path=self.path_database, mode='db')
        json_stats.update(status)
        status = self.check_file_type(file_path=self.path_query, mode='q')
        json_stats.update(status)

        # Match
        be = BELargeScale(self.path, text_sparse=512)
        if self.match_mode == 'image':
            be = BELargeScale(self.path, img_dim=True, query_batch_size=self.query_batch_size)
        elif self.match_mode == 'text_dense':
            be = BELargeScale(self.path, text_dense=True)

        start = perf_counter()
        path_match_result = self.path / 'result_match'
        make_dir(path_match_result)
        num_cat = len(json_stats['q_col_category'])
        for idx, cat in enumerate(json_stats['q_col_category']):
            # filter cat
            file_name = path_match_result / f'{cat}.{export_type}'

            # read chunk cat
            query = f"""
            select * 
            from read_parquet('{json_stats[f'db_file_clean_path']}') 
            where db_{self.col_category} = '{cat}'
            """
            chunk_db = duckdb.sql(query).pl()
            query = f"""
            select * 
            from read_parquet('{json_stats[f'q_file_clean_path']}') 
            where q_{self.col_category} = '{cat}'
            """
            chunk_q = duckdb.sql(query).pl()
            print(f"🐋 Start matching by [{self.match_mode}] cat: {cat} {idx}/{num_cat} - "
                  f"Database shape {chunk_db.shape}, Query shape {chunk_q.shape}")

            # check
            if chunk_q.shape[0] == 0 or chunk_db.shape[0] == 0:
                print(f'Database/Query have no data')
                continue
            elif file_name.exists():
                print(f'File already exists: {file_name}')
                continue

            # match
            df_match = be.match(chunk_db, chunk_q, top_k=top_k)

            # export
            if export_type == 'parquet':
                df_match.write_parquet(file_name)
            elif export_type == 'csv':
                df_match.write_csv(file_name)

            # remove caches
            rm_all_folder(self.path / 'index')
            rm_all_folder(self.path / 'result')
            for i in ['db', 'q']:
                rm_all_folder(self.path / f'{i}_array')
                rm_all_folder(self.path / f'{i}_ds')

        # update log
        time_perf = perf_counter() - start
        json_stats.update({'time_perf': time_perf, 'path result': path_match_result})
        print(f'🐋 Your files are ready, please find here: {self.path}')
        return json_stats
